from qutip.qip.device.processor import Processor
from qutip.qip.device.spinchain import (
    SpinChain, LinearSpinChain, CircularSpinChain)
from qutip.qip.device.cavityqed import DispersiveCavityQED
from qutip.qip.device.optpulseprocessor import OptPulseProcessor
from qutip.qip.device.modelprocessor import ModelProcessor
