# This is used for the deprecation.
# To remove the warning, and finalize the deprecation
# simply delete the file qutip.qip.qip_deprecation.py
# and the line bellow.
from qutip.qip.qip_deprecation import *
